"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""kv cache module"""

import math
import numpy as np
import jax
from jax import numpy as jnp
from jax.sharding import Mesh, NamedSharding, PartitionSpec as P
import queue
from inference.nn import KVCache
from transformers import PretrainedConfig

_GB = 1024**3


class KVCacheStorage:

  def __init__(
      self,
      mesh: Mesh,
      model_config: PretrainedConfig,
      page_size: int,
      hbm_utilization: float,
  ):
    """Initializes the KVCacheStorage."""
    self._mesh = mesh
    self._num_layers = model_config.num_hidden_layers
    self._num_kv_heads = model_config.num_key_value_heads
    self._head_dim = model_config.head_dim
    self._page_size = page_size
    self._hbm_kv_caches = self._init_hbm_storage(jnp.bfloat16, hbm_utilization)
    self._num_hbm_pages_per_layer = self._hbm_kv_caches[0].k.shape[1]

  @property
  def hbm_kv_caches(self) -> list[KVCache]:
    """Returns the kv cache storage."""
    return self._hbm_kv_caches

  @hbm_kv_caches.setter
  def hbm_kv_caches(self, kv_caches: list[KVCache]):
    """Sets the kv cache storage."""
    self._hbm_kv_caches = kv_caches

  @property
  def num_hbm_pages_per_layer(self) -> int:
    """Returns the number of HBM pages per layer."""
    return self._num_hbm_pages_per_layer

  def _all_devices_hbm_bytes(self, hbm_utilization: float) -> int:
    """Returns the total usable HBM bytes across all devices."""
    assert 0.0 < hbm_utilization < 1.0
    print("per device memory stats:", end="")
    try:
      per_device_memory_stats = jax.devices()[0].memory_stats()
      limit = per_device_memory_stats["bytes_reservable_limit"]
      used = per_device_memory_stats["bytes_in_use"]
      usable = int(limit * hbm_utilization) - used
      limit_GB, used_GB, usable_GB = limit // _GB, used // _GB, usable // _GB
      print(f" limit={limit_GB}GB, used={used_GB}GB, usable={usable_GB}GB")
      return usable * self._mesh.devices.size
    except:
      print(" not available")
      return 0

  def _kv_bytes_per_page(self, dtype: jnp.dtype) -> int:
    """Returns the number of bytes per key _and_ value page."""
    item_size = np.dtype(dtype).itemsize
    per_kv_bytes = self._num_kv_heads * self._head_dim * item_size * 2
    return self._page_size * per_kv_bytes

  def _shape(self) -> tuple[int, int, int, int]:
    """Returns the shape of the kv cache."""
    return (
        self._num_kv_heads,
        self.num_pages_per_layer,
        self._page_size,
        self._head_dim,
    )

  def _sharding(self) -> NamedSharding:
    """Returns the sharding of the kv cache."""
    return NamedSharding(self._mesh, P(self._mesh.axis_names, None, None, None))

  def _gen_kv_cache(self, dtype: jnp.dtype) -> KVCache:
    """Generates one per-token kv cache item."""
    # TODO: support 2d sharding
    return KVCache(
        k=jnp.ones(shape=self._shape(), device=self._sharding(), dtype=dtype),
        v=jnp.ones(shape=self._shape(), device=self._sharding(), dtype=dtype),
    )

  def _init_hbm_storage(
      self,
      dtype: jnp.dtype,
      hbm_utilization: float,
  ) -> list[KVCache]:
    """Initializes the kv cache storage across all devices."""
    all_devices_hbm_bytes = self._all_devices_hbm_bytes(hbm_utilization)
    if not all_devices_hbm_bytes >= 1 * _GB:
      raise ValueError("Insufficient HBM memory")

    kv_bytes_per_page = self._kv_bytes_per_page(dtype)
    per_layer_hbm_bytes = all_devices_hbm_bytes // self._num_layers
    self.num_pages_per_layer = per_layer_hbm_bytes // kv_bytes_per_page

    kv_storage = [self._gen_kv_cache(dtype) for _ in range(self._num_layers)]
    return kv_storage


class KVCacheManager:
  """Logical KV Cache Manager"""

  def __init__(
      self,
      num_hbm_pages,
      page_size,
  ):
    """Initializes the KVCacheManager."""
    self._page_size = page_size
    self._dummy_page_idx = 0
    self._available_hbm_pages = queue.SimpleQueue()
    for p in range(1, num_hbm_pages):
      self._available_hbm_pages.put_nowait(p)

  @property
  def page_size(self):
    """Returns the page size in the number of per-token kv cache items."""
    return self._page_size

  @property
  def dummy_page_idx(self):
    """Returns the dummy page index (0)."""
    return self._dummy_page_idx

  def alloc_prefill_hbm_pages(self, prompt_len) -> list[int]:
    """Allocates HBM pages for prompt prefill."""
    n = math.ceil(prompt_len / self._page_size)
    return self.alloc_hbm_pages(n)

  def alloc_hbm_pages(self, n: int) -> list[int]:
    """Allocates `n` HBM pages."""
    if 0 < n <= self._available_hbm_pages.qsize():
      return [self._available_hbm_pages.get(block=True) for _ in range(n)]
    else:
      return []

  def free_hbm_pages(self, pages: list[int]):
    """Frees the given HBM pages."""
    for p in pages:
      if p != self._dummy_page_idx:
        self._available_hbm_pages.put_nowait(p)



# Copyright 2025 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO(rupliu): This is a copied version of
# https://github.com/jax-ml/jax/blob/main/jax/experimental/pallas/ops/tpu/ragged_paged_attention.py
# Clean up this file when Maxtext migrate to latext Jax version

"""TPU-Friendly Ragged Paged Attention kernel.

This kernel offers a highly optimized implementation of ragged paged attention,
specifically designed for TPU and compatible with a wide range of model
specifications. It supports mixed prefill and decoding, enhancing throughput
during inference.
"""

import functools
import jax
from jax import lax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu
import jax.numpy as jnp

DEFAULT_MASK_VALUE = -0.7 * float(jnp.finfo(jnp.dtype("float32")).max)


class MultiPageAsyncCopyDescriptor:
  """Descriptor for async copy of multiple K/V pages from HBM."""

  def __init__(
      self,
      pages_hbm_ref,  # [total_num_pages, page_size, num_kv_heads_per_blk, head_dim]
      vmem_buf,  # [num_kv_pages_per_blk, page_size, num_kv_heads_per_blk, head_dim]
      sem,
      page_indices_ref,  # i32[max_num_seqs, pages_per_seq]
      offset,  # [seq_idx, kv_pages_start]
  ):
    self._vmem_buf = vmem_buf
    seq_id, kv_pages_start = offset
    self._async_copies = [
        pltpu.make_async_copy(
            pages_hbm_ref.at[page_indices_ref[seq_id, kv_pages_start + i]],
            vmem_buf.at[i],
            sem,
        )
        for i in range(vmem_buf.shape[0])
    ]

  def start(self):
    """Starts the async copies."""
    for async_copy in self._async_copies:
      async_copy.start()

  def wait(self):
    for async_copy in self._async_copies:
      async_copy.wait()
    return self._vmem_buf


def ref_ragged_paged_attention(
    queries: jax.Array,  # [max_num_batched_tokens, num_q_heads, head_dim]
    k_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    v_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    kv_lens: jax.Array,  # i32[max_num_seqs]
    page_indices: jax.Array,  # i32[max_num_seqs, pages_per_seq]
    cu_q_lens: jax.Array,  # i32[max_num_seqs + 1]
    num_seqs: jax.Array,  # i32[1],
    *,
    sm_scale: float = 1.0,
    mask_value: float = DEFAULT_MASK_VALUE,
):
  _, _, num_kv_heads, head_dim = k_pages.shape
  num_q_heads = queries.shape[1]
  assert num_q_heads % num_kv_heads == 0
  num_query_per_kv = num_q_heads // num_kv_heads
  outputs = []
  for i in range(num_seqs[0]):
    q_start = cu_q_lens[i]
    q_end = cu_q_lens[i + 1]
    q_len = q_end - q_start
    kv_len = kv_lens[i]
    indices = page_indices[i]
    q = queries[q_start:q_end]
    k = k_pages[indices, :, :, :].reshape(-1, num_kv_heads, head_dim)[:kv_len]
    v = v_pages[indices, :, :, :].reshape(-1, num_kv_heads, head_dim)[:kv_len]
    k = jnp.repeat(k, num_query_per_kv, axis=1)
    v = jnp.repeat(v, num_query_per_kv, axis=1)
    attn = jnp.einsum("qhd,khd->hqk", q, k, preferred_element_type=jnp.float32)
    attn *= sm_scale
    q_span = (kv_len - q_len) + jax.lax.broadcasted_iota(jnp.int32, attn.shape, 1)
    kv_span = jax.lax.broadcasted_iota(jnp.int32, attn.shape, 2)
    attn += jnp.where(q_span < kv_span, mask_value, 0.0)
    attn = jax.nn.softmax(attn, axis=-1).astype(v.dtype)
    out = jnp.einsum("hqk,khd->qhd", attn, v).astype(queries.dtype)
    outputs.append(out)

  return jnp.concatenate(outputs, axis=0)


# Expect to run these checkes during runtime.
def validate_inputs_on_runtime(
    q: jax.Array,  # [max_num_batched_tokens, num_q_heads, head_dim]
    k_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    v_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    kv_lens: jax.Array,  # i32[max_num_seqs]
    page_indices: jax.Array,  # i32[max_num_seqs, pages_per_seq]
    cu_q_lens: jax.Array,  # i32[max_num_seqs + 1]
    num_seqs,  # i32[1]
):
  check_inputs_shapes(q, k_pages, v_pages, kv_lens, page_indices, cu_q_lens, num_seqs)
  max_num_batched_tokens = q.shape[0]
  page_size = k_pages.shape[1]
  max_num_seqs, pages_per_seq = page_indices.shape
  if num_seqs[0] > max_num_seqs:
    raise ValueError(f"{num_seqs[0]=} must be less or equal to {max_num_seqs=}")
  max_kv_len = jnp.max(kv_lens)
  min_pages_per_seq = ceil_div(max_kv_len, page_size)
  if pages_per_seq < min_pages_per_seq:
    raise ValueError(
        f"{pages_per_seq=} must be greater or equal to" f" {min_pages_per_seq=} given {max_kv_len=} and {page_size=}."
    )
  if cu_q_lens[num_seqs[0]] > max_num_batched_tokens:
    raise ValueError(f"Total q tokens {cu_q_lens[num_seqs[0]]} must be less or equal to" f" {max_num_batched_tokens=}.")
  for i in range(num_seqs[0]):
    q_len = cu_q_lens[i + 1] - cu_q_lens[i]
    kv_len = kv_lens[i]
    if q_len > kv_len:
      raise ValueError(f"{q_len=} must be less or equal to {kv_len=} at sequence {i}.")


# Expect to run these checks during compile time.
def check_inputs_shapes(
    q: jax.Array,  # [max_num_batched_tokens, num_q_heads, head_dim]
    k_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    v_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    kv_lens: jax.Array,  # i32[max_num_seqs]
    page_indices: jax.Array,  # i32[max_num_seqs, pages_per_seq]
    cu_q_lens: jax.Array,  # i32[max_num_seqs + 1]
    num_seqs,  # i32[1]
):
  _, num_q_heads, head_dim = q.shape
  _, _, num_kv_heads, head_dim_k = k_pages.shape
  max_num_seqs, _ = page_indices.shape
  if num_seqs.shape != (1,):
    raise ValueError(f"{num_seqs.shape=} must be (1,)")
  if k_pages.shape != v_pages.shape:
    raise ValueError(f"{k_pages.shape=} and {v_pages.shape=} must have the same shape.")
  if head_dim_k != head_dim:
    raise ValueError(f"Q head_dim {head_dim} must be the same as that of K/V {head_dim_k}.")
  if kv_lens.shape != (max_num_seqs,):
    raise ValueError(
        f"Expected {kv_lens.shape=} to be ({max_num_seqs},) where" " `max_num_seqs` is `page_indices.shape[0]`."
    )
  if cu_q_lens.shape != (max_num_seqs + 1,):
    raise ValueError(
        f"Expected {cu_q_lens.shape=} to be ({max_num_seqs + 1},)  where" " `max_num_seqs` is `page_indices.shape[0]`."
    )
  if kv_lens.dtype != jnp.int32 or page_indices.dtype != jnp.int32 or cu_q_lens.dtype != jnp.int32:
    raise ValueError(
        "The dtype of `kv_lens`, `page_indices`, and `cu_q_lens` must be"
        f" int32. Got {kv_lens.dtype=}, {page_indices.dtype=},"
        f" {cu_q_lens.dtype=}."
    )
  if num_q_heads % num_kv_heads != 0:
    raise ValueError(f"{num_q_heads=} must be divisible by {num_kv_heads=}")


def ragged_paged_attention_kernel(
    # Prefetch
    kv_lens_ref,  # [max_num_seqs]
    page_indices_ref,  # [max_num_seqs, pages_per_seq]
    cu_q_lens_ref,  # [max_num_seqs + 1]
    seq_buf_idx_ref,
    # TODO(jevinjiang): if OOM in SMEM, consider pack to other scalar refs.
    num_seqs_ref,
    # Input
    q_ref,  # [num_q_per_blk, num_q_heads_per_blk, head_dim]
    k_pages_hbm_ref,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    v_pages_hbm_ref,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    # Output
    o_ref,  # [num_q_per_blk, num_q_heads_per_blk, head_dim]
    # Scratch
    k_bufs,  # [2, num_kv_pages_per_blk, page_size, num_kv_heads_per_blk, head_dim]
    v_bufs,  # [2, num_kv_pages_per_blk, page_size, num_kv_heads_per_blk, head_dim]
    sems,  # [2, 2]
    l_ref,  # [num_kv_heads_per_blk, num_q_per_blk * num_q_heads_per_kv_head, 128]
    m_ref,  # [num_kv_heads_per_blk, num_q_per_blk * num_q_heads_per_kv_head, 128]
    *,
    sm_scale: float,
    mask_value: float,
):
  num_q_per_blk, num_q_heads_per_blk, head_dim = q_ref.shape
  num_seqs = num_seqs_ref[0]
  _, num_kv_pages_per_blk, page_size, num_kv_heads_per_blk, _ = k_bufs.shape
  num_kv_per_blk = num_kv_pages_per_blk * page_size
  num_q_heads_per_kv_head = num_q_heads_per_blk // num_kv_heads_per_blk
  heads_blk_idx, q_blk_idx = (
      pl.program_id(0),
      pl.program_id(1),
  )
  num_heads_blks = pl.num_programs(0)
  init_seq_idx = seq_buf_idx_ref[0]
  init_buf_idx = seq_buf_idx_ref[1]
  q_len_start = q_blk_idx * num_q_per_blk
  q_len_end = q_len_start + num_q_per_blk

  def create_kv_async_copy_descriptors(heads_blk_idx, seq_idx, kv_blk_idx, buf_idx):
    offset = (seq_idx, kv_blk_idx * num_kv_pages_per_blk)
    heads_start = heads_blk_idx * num_kv_heads_per_blk
    async_copy_k = MultiPageAsyncCopyDescriptor(
        k_pages_hbm_ref.at[:, :, pl.ds(heads_start, num_kv_heads_per_blk), :],
        k_bufs.at[buf_idx],
        sems.at[buf_idx, 0],
        page_indices_ref,
        offset,
    )
    async_copy_v = MultiPageAsyncCopyDescriptor(
        v_pages_hbm_ref.at[:, :, pl.ds(heads_start, num_kv_heads_per_blk), :],
        v_bufs.at[buf_idx],
        sems.at[buf_idx, 1],
        page_indices_ref,
        offset,
    )
    return async_copy_k, async_copy_v

  # TODO(jevinjiang): Add these to Mosaic:
  # 1. Support arbitrary strided load/store for any dtype.
  # 2. Support arbitrary strided load/store for any last dimension.
  def strided_load_kv(ref, start, step):
    if ref.dtype == jnp.float32:
      return ref[start::step, :]
    packing = get_dtype_packing(ref.dtype)
    assert ref.dtype == jnp.bfloat16
    assert step % packing == 0
    b_start = start // packing
    b_offset = start % packing
    b_step = step // packing
    b_ref = ref.bitcast(jnp.int32)
    b = b_ref[b_start::b_step, :]
    bw = 32 // packing
    b = jnp.right_shift(b, bw * b_offset)
    b = jnp.left_shift(b, bw * (packing - 1))
    return pltpu.bitcast(b, jnp.float32).astype(jnp.bfloat16)

  def fold_on_2nd_minor(vec):
    assert vec.dtype == jnp.bfloat16 or vec.dtype == jnp.float32
    assert len(vec.shape) >= 2
    last_dim = vec.shape[-1]
    packing = get_dtype_packing(vec.dtype)
    if vec.shape[-2] % packing != 0:
      vec = vec.astype(jnp.float32)
    return vec.reshape(-1, last_dim)

  @pl.when(heads_blk_idx + q_blk_idx == 0)
  def prefetch_first_kv_blk():
    async_copy_k, async_copy_v = create_kv_async_copy_descriptors(heads_blk_idx, init_seq_idx, 0, init_buf_idx)
    async_copy_k.start()
    async_copy_v.start()

  def is_cur_q_blk_needed(q_states):
    done, cur_seq_idx, _ = q_states
    return jnp.logical_and(done == 0, cur_seq_idx < num_seqs)

  def compute_with_cur_q_blk(q_states):
    done, cur_seq_idx, cur_buf_idx = q_states
    q_start = cu_q_lens_ref[cur_seq_idx]
    q_end = cu_q_lens_ref[cur_seq_idx + 1]
    q_len = q_end - q_start
    kv_len = kv_lens_ref[cur_seq_idx]

    def get_next_prefetch_ids(heads_blk_idx, cur_seq_idx, kv_blk_idx, cur_buf_idx):
      next_kv_blk_idx = kv_blk_idx + 1
      is_last_kv_blk = next_kv_blk_idx * num_kv_per_blk >= kv_len
      next_kv_blk_idx = lax.select(
          is_last_kv_blk,
          0,
          next_kv_blk_idx,
      )
      is_cur_seq_end_in_cur_q_blk = q_end <= q_len_end
      next_seq_idx = lax.select(
          is_last_kv_blk,
          lax.select(is_cur_seq_end_in_cur_q_blk, cur_seq_idx + 1, cur_seq_idx),
          cur_seq_idx,
      )
      is_last_seq = next_seq_idx == num_seqs
      next_seq_idx = lax.select(
          is_last_seq,
          0,
          next_seq_idx,
      )
      next_heads_blk_idx = lax.select(
          is_last_seq,
          heads_blk_idx + 1,
          heads_blk_idx,
      )
      next_buf_idx = lax.select(cur_buf_idx == 0, 1, 0)
      return next_heads_blk_idx, next_seq_idx, next_kv_blk_idx, next_buf_idx

    def flash_attention(
        q,  # [num_q_per_blk * num_q_heads_per_kv_head, head_dim]
        k,  # [num_kv_per_blk, head_dim]
        v,  # [num_kv_per_blk, head_dim]
        head_l_ref,  # [num_q_per_blk * num_q_heads_per_kv_head, 128]
        head_m_ref,  # [num_q_per_blk * num_q_heads_per_kv_head, 128]
        head_o_ref,  # [num_q_per_blk, num_q_heads_per_kv_head, head_dim]
        *,
        kv_blk_idx,
    ):
      assert q.shape == (
          num_q_per_blk * num_q_heads_per_kv_head,
          head_dim,
      )
      assert k.shape == (
          num_kv_per_blk,
          head_dim,
      ), f"{k.shape=}, {(num_kv_per_blk, head_dim)=} {k.dtype=}"
      assert v.shape == (num_kv_per_blk, head_dim)
      assert head_m_ref.shape == (
          num_q_per_blk * num_q_heads_per_kv_head,
          128,
      )
      assert head_l_ref.shape == (
          num_q_per_blk * num_q_heads_per_kv_head,
          128,
      )
      assert head_o_ref.shape == (
          num_q_per_blk,
          num_q_heads_per_kv_head,
          head_dim,
      )
      kv_len_start = kv_blk_idx * num_kv_per_blk

      def masked_store(ref, val, start, end, group=1):
        iota = lax.broadcasted_iota(jnp.int32, ref.shape, 0) // group
        mask = jnp.logical_and(iota >= start, iota < end)
        pl.store(ref, tuple(slice(None) for _ in ref.shape), val, mask=mask)

      qk = jnp.einsum("nd,md->nm", q, k, preferred_element_type=jnp.float32) * sm_scale
      store_start = jnp.maximum(q_start - q_len_start, 0)
      store_end = jnp.minimum(q_end - q_len_start, num_q_per_blk)

      @pl.when(kv_blk_idx == 0)
      def init_scratch_ref():
        masked_store(
            head_m_ref,
            jnp.full_like(head_m_ref, -jnp.inf),
            store_start,
            store_end,
            num_q_heads_per_kv_head,
        )
        masked_store(
            head_l_ref,
            jnp.zeros_like(head_l_ref),
            store_start,
            store_end,
            num_q_heads_per_kv_head,
        )
        masked_store(
            head_o_ref,
            jnp.zeros_like(head_o_ref),
            store_start,
            store_end,
        )

      row_ids = (
          (kv_len - q_len)
          + q_len_start
          - q_start
          + jax.lax.broadcasted_iota(
              jnp.int32,
              (num_q_per_blk * num_q_heads_per_kv_head, num_kv_per_blk),
              0,
          )
          // num_q_heads_per_kv_head
      )
      col_ids = kv_len_start + jax.lax.broadcasted_iota(
          jnp.int32,
          (num_q_per_blk * num_q_heads_per_kv_head, num_kv_per_blk),
          1,
      )
      causal_mask = row_ids < col_ids
      qk += jnp.where(causal_mask, mask_value, 0.0)
      m_curr = jnp.max(qk, axis=1, keepdims=True)
      s_curr = jnp.exp(qk - m_curr)
      qkv = jnp.dot(s_curr, v, preferred_element_type=jnp.float32)
      lm_store_shape = head_m_ref.shape
      m_curr = jnp.broadcast_to(m_curr, lm_store_shape)
      l_curr = jnp.broadcast_to(s_curr.sum(axis=1, keepdims=True), lm_store_shape)
      m_prev = head_m_ref[...]
      l_prev = head_l_ref[...]
      m_next = jnp.maximum(m_prev, m_curr)
      masked_store(head_m_ref, m_next, store_start, store_end, num_q_heads_per_kv_head)
      alpha = jnp.exp(m_prev - m_next)
      beta = jnp.exp(m_curr - m_next)
      l_alpha = alpha * l_prev
      l_next = l_alpha + beta * l_curr
      l_next_safe = jnp.where(l_next == 0.0, 1.0, l_next)
      masked_store(
          head_l_ref,
          l_next_safe,
          store_start,
          store_end,
          num_q_heads_per_kv_head,
      )

      def broadcast_to_shape(arr, shape):
        if arr.shape == shape:
          return arr
        assert len(arr.shape) == len(shape)
        assert arr.shape[0] == shape[0]
        assert shape[1] % arr.shape[1] == 0
        # no-op concatenation.
        return jnp.concatenate([arr for _ in range(shape[1] // arr.shape[1])], axis=1)

      o_curr = head_o_ref[...].reshape(-1, head_dim)
      l_alpha = broadcast_to_shape(l_alpha, qkv.shape)
      beta = broadcast_to_shape(beta, qkv.shape)
      l_next_safe = broadcast_to_shape(l_next_safe, qkv.shape)
      out = lax.div(
          l_alpha * o_curr + beta * qkv,
          l_next_safe,
      ).astype(head_o_ref.dtype)
      masked_store(
          head_o_ref,
          out.reshape(head_o_ref.shape),
          store_start,
          store_end,
      )

    def is_valid_kv_blk_in_cur_seq(kv_states):
      kv_blk_idx, _ = kv_states
      return kv_blk_idx * num_kv_per_blk < kv_len

    def compute_with_kv_blk_in_cur_seq(kv_states):
      kv_blk_idx, cur_buf_idx = kv_states
      next_heads_blk_idx, next_seq_idx, next_kv_blk_idx, next_buf_idx = get_next_prefetch_ids(
          heads_blk_idx, cur_seq_idx, kv_blk_idx, cur_buf_idx
      )

      @pl.when(next_heads_blk_idx < num_heads_blks)
      def prefetch_next_kv_blk():
        # TODO(jevinjiang): reuse the same buffer if it is already prefetched!
        # TODO(jevinjiang): only fetch effective dynamic size to hold kv_len and
        # DMA to fixed size buffer!
        next_async_copy_k, next_async_copy_v = create_kv_async_copy_descriptors(
            next_heads_blk_idx, next_seq_idx, next_kv_blk_idx, next_buf_idx
        )
        next_async_copy_k.start()
        next_async_copy_v.start()

      cur_async_copy_k, cur_async_copy_v = create_kv_async_copy_descriptors(
          heads_blk_idx, cur_seq_idx, kv_blk_idx, cur_buf_idx
      )
      kv_to_load_shape = (
          num_kv_pages_per_blk * page_size * num_kv_heads_per_blk,
          head_dim,
      )
      k_ref = cur_async_copy_k.wait().reshape(kv_to_load_shape)
      v_ref = cur_async_copy_v.wait().reshape(kv_to_load_shape)
      for kv_head_idx in range(num_kv_heads_per_blk):
        q_head_idx = kv_head_idx * num_q_heads_per_kv_head
        # TODO(jevinjiang): extra handlig for packed type that can start at
        # unaligned position!
        q = fold_on_2nd_minor(q_ref[:, q_head_idx : q_head_idx + num_q_heads_per_kv_head, :])
        k = strided_load_kv(k_ref, kv_head_idx, num_kv_heads_per_blk)
        v = strided_load_kv(v_ref, kv_head_idx, num_kv_heads_per_blk)
        flash_attention(
            q,
            k,
            v,
            l_ref.at[kv_head_idx],
            m_ref.at[kv_head_idx],
            o_ref.at[:, q_head_idx : q_head_idx + num_q_heads_per_kv_head, :],
            kv_blk_idx=kv_blk_idx,
        )
      return kv_blk_idx + 1, next_buf_idx

    _, next_buf_idx = lax.while_loop(
        is_valid_kv_blk_in_cur_seq,
        compute_with_kv_blk_in_cur_seq,
        (0, cur_buf_idx),  # (kv_blk_idx, buf_idx)
    )
    next_seq_idx = lax.select(q_end <= q_len_end, cur_seq_idx + 1, cur_seq_idx)
    done = lax.select(q_end < q_len_end, done, 1)
    return done, next_seq_idx, next_buf_idx

  _, seq_idx, buf_idx = lax.while_loop(
      is_cur_q_blk_needed,
      compute_with_cur_q_blk,
      (0, init_seq_idx, init_buf_idx),  # (done, seq_idx, buf_idx)
  )
  # Reset seq_idx for next kv_heads_blk if run out of seqs!
  seq_buf_idx_ref[0] = lax.select(seq_idx < num_seqs, seq_idx, 0)
  seq_buf_idx_ref[1] = buf_idx


def ceil_div(a, b):
  assert b != 0
  return (a + b - 1) // b


def get_dtype_packing(dtype):
  if dtype == jnp.float32:
    return 1
  if dtype == jnp.bfloat16:
    return 2
  if dtype == jnp.int8:
    return 4
  if dtype == jnp.int4:
    return 8
  raise ValueError(f"Not implemented: unsupported {dtype=}")


def get_min_heads_per_blk(num_q_heads, num_kv_heads, q_dtype, kv_dtype):
  q_packing = get_dtype_packing(q_dtype)
  kv_packing = get_dtype_packing(kv_dtype)

  def can_be_xla_fully_tiled(x, packing):
    if x % packing != 0:
      return False
    x //= packing
    return x in (1, 2, 4, 8) or x % 8 == 0

  # TODO(jevinjiang): support unaligned number of heads!
  if not can_be_xla_fully_tiled(num_kv_heads, kv_packing):
    raise ValueError(f"Not implemented: {num_kv_heads=} can not be XLA fully tiled.")
  assert num_q_heads % num_kv_heads == 0
  ratio = num_q_heads // num_kv_heads
  # TODO(jevinjiang): we can choose smaller tiling for packed type if large
  # second minor tiling is not on.
  max_kv_tiling = 8 * kv_packing
  min_kv_heads = max_kv_tiling if num_kv_heads % max_kv_tiling == 0 else num_kv_heads
  min_q_heads = min_kv_heads * ratio
  if can_be_xla_fully_tiled(min_q_heads, q_packing):
    return min_q_heads, min_kv_heads
  return num_q_heads, num_kv_heads


@functools.partial(
    jax.jit,
    static_argnames=[
        "sm_scale",
        "mask_value",
        "num_kv_pages_per_block",
        "num_queries_per_block",
        "vmem_limit_bytes",
    ],
)
def ragged_paged_attention(
    q: jax.Array,  # [max_num_batched_tokens, num_q_heads, head_dim]
    # TODO(jevinjiang): create a write_to_kv_cache kernel!
    k_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    v_pages: jax.Array,  # [total_num_pages, page_size, num_kv_heads, head_dim]
    kv_lens: jax.Array,  # i32[max_num_seqs]
    page_indices: jax.Array,  # i32[max_num_seqs, pages_per_seq]
    cu_q_lens: jax.Array,  # i32[max_num_seqs + 1]
    num_seqs: jax.Array,  # i32[1]
    *,
    sm_scale: float = 1.0,
    mask_value: float = DEFAULT_MASK_VALUE,
    num_kv_pages_per_block: int = 16,
    num_queries_per_block: int = 128,
    vmem_limit_bytes: int | None = None,
):
  """Ragged paged attention that supports mixed prefill and decode.

  Args:
    q: concatenated all sequences' queries.
    k_pages: paged K cache. Normally in HBM.
    v_pages: paged V cache. Normally in HBM.
    kv_lens: padded kv lengths. Only the first num_seqs values are valid.
    page_indices: the first index indicates which page to use in the kv cache
      for each sequence. Only the first num_seqs values are valid.
    cu_q_lens: the cumulative sum of the effective query lengths. Similar to
      kv_lens, only the first num_seqs+1 values are valid.
    num_seqs: the dynamic number of sequences.
    sm_scale: the softmax scale which will be applied to the Q@K^T.
    mask_value: mask value for causal mask.
    num_kv_pages_per_block: number of kv pages to be processed in one flash
      attention block in the pallas kernel.
    num_queries_per_block: number of kv pages to be processed in one flash
      attention block in the pallas kernel.
    vmem_limit_bytes: the vmem limit for the pallas kernel.

  Returns:
    The output of the attention.
  """
  check_inputs_shapes(q, k_pages, v_pages, kv_lens, page_indices, cu_q_lens, num_seqs)
  _, num_q_heads, head_dim = q.shape
  _, page_size, num_kv_heads, _ = k_pages.shape
  num_q_per_blk = num_queries_per_block
  num_kv_pages_per_blk = num_kv_pages_per_block
  num_q_heads_per_kv_head = num_q_heads // num_kv_heads
  num_q_blks = ceil_div(cu_q_lens[num_seqs[0]], num_q_per_blk)
  num_q_heads_per_blk, num_kv_heads_per_blk = get_min_heads_per_blk(num_q_heads, num_kv_heads, q.dtype, k_pages.dtype)
  assert num_q_heads_per_blk % num_q_heads_per_kv_head == 0
  num_heads_blks = num_q_heads // num_q_heads_per_blk
  grid = (num_heads_blks, num_q_blks)

  def q_index_map(heads_blk_idx, q_blk_idx, *_):
    return (q_blk_idx, heads_blk_idx, 0)

  q_block_spec = pl.BlockSpec(
      (num_q_per_blk, num_q_heads_per_blk, head_dim),
      q_index_map,
  )
  in_specs = [
      q_block_spec,
      pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
      pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
  ]
  out_specs = q_block_spec
  lm_scratch = pltpu.VMEM(
      # TODO(jevinjiang): use 128 instead of 1 is due to Mosaic does not support
      # unaligned slicing!
      (num_kv_heads_per_blk, num_q_per_blk * num_q_heads_per_kv_head, 128),
      jnp.float32,
  )
  double_buf_scratch = pltpu.VMEM(
      (
          2,  # For double buffering during DMA copies.
          num_kv_pages_per_blk,
          page_size,
          num_kv_heads_per_blk,
          head_dim,
      ),
      k_pages.dtype,
  )
  scratch_shapes = [
      double_buf_scratch,  # k_bufs
      double_buf_scratch,  # v_bufs
      pltpu.SemaphoreType.DMA((2, 2)),  # [double_buffers, k_sem/v_sem]
      lm_scratch,  # l_ref
      lm_scratch,  # m_ref
  ]
  scalar_prefetches = (
      kv_lens,
      page_indices,
      cu_q_lens,
      jnp.array((0, 0), jnp.int32),  # seq_idx, buf_idx
      num_seqs,
  )
  kernel = pl.pallas_call(
      functools.partial(
          ragged_paged_attention_kernel,
          sm_scale=sm_scale,
          mask_value=mask_value,
      ),
      grid_spec=pltpu.PrefetchScalarGridSpec(
          num_scalar_prefetch=len(scalar_prefetches),
          in_specs=in_specs,
          out_specs=out_specs,
          grid=grid,
          scratch_shapes=scratch_shapes,
      ),
      compiler_params=pltpu.TPUCompilerParams(
          dimension_semantics=(
              "arbitrary",
              "arbitrary",
          ),
          vmem_limit_bytes=vmem_limit_bytes,
      ),
      out_shape=jax.ShapeDtypeStruct(shape=q.shape, dtype=jnp.float32),
      name="ragged_paged_attention_kernel",
  )
  # TODO(jevinjiang): Use f32 acc scratch for output! So we only need
  # to transfer output with desired dtype back to HBM.
  return kernel(*scalar_prefetches, q, k_pages, v_pages).astype(q.dtype)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# from absl.testing import absltest
# import numpy as np
# import jax.experimental
# import jax.experimental.mesh_utils
# import jax
# from jax import numpy as jnp
# from jax._src import test_util as jtu
# from jax.sharding import NamedSharding, PartitionSpec as P
# from inference import nn
# from inference import parallel

# """TODO: Enable the attention test"""


# class AttentionTest(absltest.TestCase):

#   def _create_device_mesh(self):
#     devices = jax.devices()
#     return parallel.create_device_mesh(
#         devices=devices,
#         shape=(len(devices), 1),
#     )

#   def test_kv_cache_sharding(self):
#     axis = parallel.tp_axis_names()
#     k_hbm = jnp.zeros((8, 8, 8, 128))
#     v_hbm = jnp.copy(k_hbm)
#     num_layer = 3
#     mesh = self._create_device_mesh
#     kv_cache = [
#         nn.KVCache(
#             k=k_hbm,
#             v=v_hbm,
#         )
#         for _ in range(num_layer)
#     ]
#     kv_cache_sharding = [
#         nn.KVCache(
#             k=NamedSharding(mesh, P(axis, None, None, None)),
#             v=NamedSharding(mesh, P(axis, None, None, None)),
#         )
#         for _ in range(num_layer)
#     ]

#     def sharding(data, sharding):
#       return jax.device_put(data, sharding)

#     kv_cache = jax.tree.map(sharding, kv_cache, kv_cache_sharding)
#     for i in range(num_layer):
#       self.assertIsInstance(kv_cache[i].k.sharding, NamedSharding)
#       self.assertIsInstance(kv_cache[i].v.sharding, NamedSharding)

#   def test_prefill(self):
#     if jtu.device_under_test() != "tpu":
#       self.skipTest("Only supports TPU")

#     num_attn_heads = 16
#     num_kv_heads = 8
#     head_dim = 128
#     total_page_num = 16
#     page_size = 8
#     attn_layer = nn.AttentionOps(num_attn_heads, num_kv_heads, head_dim)
#     k_hbm, v_hbm = (
#         jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
#         jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
#     )

#     kv_cache = nn.KVCache(k=k_hbm, v=v_hbm)

#     prefill_len = 16
#     prefill_non_padding_len = 12
#     q = jnp.ones(((prefill_len, num_attn_heads, head_dim)))
#     k_to_save, v_to_save = (
#         jnp.ones((prefill_len, num_kv_heads, head_dim)),
#         jnp.ones((prefill_len, num_kv_heads, head_dim)),
#     )
#     num_page_to_use = prefill_len // page_size
#     # using the second and third page to save the kv cache.
#     page_table = jnp.array([1, 3, 0, 0, 0, 0])
#     output, kv_cache = attn_layer._prefill(
#         q,
#         k_to_save,
#         v_to_save,
#         kv_cache,
#         nn.AttentionMetadata(
#             prefill_length=prefill_non_padding_len,
#             prefill_pos=jnp.arange(0, 8),
#             prefill_page_table=page_table,
#             generate_pos=0,
#             generate_page_table=0,
#         ),
#     )
#     kv_cache = attn_layer._write_prefill_kv_to_kv_cache(
#         k_to_save, v_to_save, kv_cache, prefill_non_padding_len, page_table
#     )
#     np.testing.assert_allclose(
#         kv_cache.k[:, page_table[:2], :, :],
#         jnp.ones((num_kv_heads, num_page_to_use, page_size, head_dim)),
#     )
#     np.testing.assert_allclose(
#         kv_cache.v[:, page_table[:2], :, :],
#         jnp.ones((num_kv_heads, num_page_to_use, page_size, head_dim)),
#     )
#     zero_index = [i for i in range(page_size)]
#     zero_index = (
#         zero_index[0 : page_table[0]]
#         + zero_index[page_table[0] + 1 : page_table[1]]
#         + zero_index[page_table[1] + 1 :]
#     )

#     np.testing.assert_allclose(
#         kv_cache.k[:, zero_index, :, :],
#         jnp.zeros(
#             (num_kv_heads, page_size - num_page_to_use, page_size, head_dim)
#         ),
#     )
#     np.testing.assert_allclose(
#         kv_cache.v[:, zero_index, :, :],
#         jnp.zeros(
#             (num_kv_heads, page_size - num_page_to_use, page_size, head_dim)
#         ),
#     )
#     np.testing.assert_equal(
#         output.shape, (prefill_len, num_attn_heads * head_dim)
#     )

#   def test_generate_cache_update(self):
#     num_attn_heads = 16
#     num_kv_heads = 8
#     head_dim = 128
#     total_page_num = 128
#     page_size = 8
#     max_len = 32
#     attn_layer = nn.AttentionOps(num_attn_heads, num_kv_heads, head_dim)
#     k_hbm, v_hbm = (
#         jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
#         jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
#     )

#     kv_cache = nn.KVCache(k=k_hbm, v=v_hbm)

#     generate_len = 4
#     k_to_save, v_to_save = (
#         jnp.ones((generate_len, num_kv_heads, head_dim)),
#         jnp.ones((generate_len, num_kv_heads, head_dim)),
#     )
#     prng = jax.random.PRNGKey(99)
#     page_table = jnp.asarray(
#         np.random.choice(
#             total_page_num, (generate_len, max_len // page_size), replace=False
#         )
#     )
#     page_pos = jax.random.randint(
#         prng, shape=(generate_len,), minval=0, maxval=total_page_num * page_size
#     )
#     kv_cache = attn_layer._write_generate_kv_to_kv_cache(
#         k_to_save, v_to_save, kv_cache, page_pos, page_table
#     )

#     page_idx, offset = jnp.divmod(page_pos, page_size)
#     page_to_update = page_table[jnp.arange(0, generate_len), page_idx]

#     np.testing.assert_allclose(
#         kv_cache.k[:, page_to_update, offset, :],
#         jnp.ones_like(kv_cache.k[:, page_to_update, offset, :]),
#     )

#     np.testing.assert_allclose(
#         kv_cache.v[:, page_to_update, offset, :],
#         jnp.ones_like(kv_cache.v[:, page_to_update, offset, :]),
#     )

#     np.testing.assert_allclose(
#         jnp.sum(kv_cache.k), generate_len * num_kv_heads * head_dim
#     )

#   def test_generate(self):
#     if jtu.device_under_test() != "tpu":
#       self.skipTest("Only supports TPU")

#     num_attn_heads = 16
#     num_kv_heads = 8
#     head_dim = 128
#     total_page_num = 128
#     page_size = 8
#     max_len = 64
#     attn_layer = nn.AttentionOps(num_attn_heads, num_kv_heads, head_dim)
#     k_hbm, v_hbm = (
#         jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
#         jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
#     )

#     kv_cache = nn.KVCache(k=k_hbm, v=v_hbm)

#     num_generate_tokens = 4
#     q = jnp.ones((num_generate_tokens, num_attn_heads, head_dim))
#     k_to_save, v_to_save = (
#         jnp.ones((num_generate_tokens, num_kv_heads, head_dim)),
#         jnp.ones((num_generate_tokens, num_kv_heads, head_dim)),
#     )
#     prng = jax.random.PRNGKey(99)
#     page_table = jnp.asarray(
#         np.random.choice(
#             total_page_num,
#             (num_generate_tokens, max_len // page_size),
#             replace=False,
#         )
#     )
#     page_pos = jax.random.randint(
#         prng,
#         shape=(num_generate_tokens,),
#         minval=0,
#         maxval=total_page_num * page_size,
#     )

#     output, kv_cache = attn_layer._generate(
#         q,
#         k_to_save,
#         v_to_save,
#         kv_cache,
#         nn.AttentionMetadata(
#             prefill_length=0,
#             prefill_pos=0,
#             prefill_page_table=0,
#             generate_pos=page_pos,
#             generate_page_table=page_table,
#         ),
#     )

#     page_idx, offset = jnp.divmod(page_pos, page_size)
#     page_to_update = page_table[jnp.arange(0, num_generate_tokens), page_idx]

#     np.testing.assert_allclose(
#         kv_cache.k[:, page_to_update, offset, :],
#         jnp.ones_like(kv_cache.k[:, page_to_update, offset, :]),
#     )

#     np.testing.assert_allclose(
#         kv_cache.v[:, page_to_update, offset, :],
#         jnp.ones_like(kv_cache.v[:, page_to_update, offset, :]),
#     )

#     np.testing.assert_allclose(
#         jnp.sum(kv_cache.k), num_generate_tokens * num_kv_heads * head_dim
#     )
#     np.testing.assert_equal(
#         output.shape, (num_generate_tokens, num_attn_heads * head_dim)
#     )


# if __name__ == "__main__":
#   absltest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metrics util classes for collecting and managing metrics."""

import datetime

import numpy as np
from typing import Tuple, List, Dict


def _floor_datetime_to_sec(timestamp: datetime.datetime) -> datetime.datetime:
  """ "Floor the timestamp to the nearest most recent second"""
  return timestamp.replace(microsecond=0)


def _now_floored_to_second() -> datetime.datetime:
  """Return the current timestamp floored to the nearest most recent second a"""
  now = datetime.datetime.now()
  return _floor_datetime_to_sec(now)


class EventMetric:
  """An event metric  for distribution stats reporting. Not thread-safe."""

  def __init__(self, name: str, description: str, unit: str = ""):
    self._name = name
    self._description = description
    self._unit = unit
    self._data = []

  def data(self) -> List[float]:
    """Returns all stored data points.

    Returns:
      A list of data points in the order that was stored
    """
    return self._data

  def record(self, value: float):
    """Record a data point

    Args:
      value: The data point to be stored.
    """
    self._data.append(value)

  def percentile(self, percentile: int) -> float:
    """Computes and returns the specified percentile of the collected data.

    Args:
      percentile: The percentile to compute.

    Returns:
      The computed percentile.
    """
    if not 0 <= percentile <= 100:
      raise ValueError(f"Percentile {percentile} is not in [0, 100]")
    if not self._data:
      raise ValueError(
          f"No data points in metric {self._name} to compute percentile"
      )
    return np.percentile(self._data, percentile)

  def mean(self) -> float:
    """Calculates and returns the mean value of the collected data.

    Returns:
        The mean value of the collected data
    """
    if not self._data:
      raise ValueError(f"No data points in metric {self._name} to compute mean")
    return np.mean(self._data)

  def distribution_summary_str(self) -> str:
    """Generates a string representation of the distribution summary

    Returns:
        The string representation of the distribution summary including
        mean, p50, p90 and p99.
    """
    s = ""
    s += f"Mean {self._name}: {self.mean():.2f} {self._unit}\n"
    s += f"Median {self._name}: {self.percentile(50):.2f} {self._unit}\n"
    s += f"P99 {self._name}: {self.percentile(99):.2f} {self._unit}"
    return s

  def distribution_summary_dict(self) -> dict[str, float]:
    """Generates a dictionary representation of the distribution summary

    Returns:
      A dictionary containing of the distribution summary including mean,
      p50, p90 and p99.
    """
    return {
        f"mean_{self._name}_{self._unit}": self.mean(),
        f"median_{self._name}_{self._unit}": self.percentile(50),
        f"p99_{self._name}_{self._unit}": self.percentile(99),
    }


class CounterMetric:
  """A count metric for computing rates over time. Not thread-safe."""

  def __init__(self, name: str, description: str):
    self._name = name
    self._description = description
    self._data: dict[datetime.datetime, int] = {}

  def data(self) -> Dict[datetime.datetime, int]:
    """Returns all stored data points.

    Returns:
      A dictionary of data points where the key is the timestamp and the value
      is the aggregated counts within the second of the timestamp.
    """
    return self._data

  def total_count(self) -> int:
    """Returns aggregated counts

    Returns:
      The aggregated counts.
    """
    return sum(self._data.values())

  def total_duration_sec(self) -> int:
    """Returns the duration between the first and last count increment

    Returns:
        The duration (in seconds) between the first and last increment
        (inclusive of both ends).
    """
    start_time = min(self._data.keys())
    end_time = max(self._data.keys())
    return int((end_time - start_time).total_seconds() + 1)

  def increment(
      self, count: int = 1, timestamp: datetime.datetime | None = None
  ):
    """Increment the counter by count

    Args:
        count: The amount to increment
        timestamp: The timestamp for the increment. Default to now if none is
          provided.
    """
    if timestamp is None:
      cur_time = _now_floored_to_second()
    else:
      cur_time = _floor_datetime_to_sec(timestamp)
    # Add timestamp with default value 0 if doesn't exist
    cur_count = self._data.setdefault(cur_time, 0)
    self._data[cur_time] = cur_count + count
    return

  def rate(self) -> float:
    """Calculates the rate of change between the first and last increments.

    Returns:
      The rate of change between the first and last increments.
    """
    if len(self._data.keys()) < 2:
      raise ValueError(
          "At least 2 data points are required to compute the rate"
      )
    start_time = min(self._data.keys())
    end_time = max(self._data.keys())
    delta_time_sec = (end_time - start_time).total_seconds()
    sorted_counts = [count for timestamp, count in sorted(self._data.items())]
    delta_count = sum(sorted_counts[1:])
    return delta_count / delta_time_sec

  def rate_over_window(
      self, window_size_sec: int
  ) -> List[Tuple[datetime.datetime, float]]:
    """Calculate the rates over time."

    Args:
      window_size_sec: The size of the window in seconds for computing each
        individual rate

    Returns:
      A list of rates over time, where each element represents the rate of
      change for the specified window size.
    """
    if len(self._data.keys()) < 2:
      raise ValueError(
          f"At least 2 different timestamp values are required to calculate "
          f"the rate, but have only {len(self._data.keys())}"
      )
    rates: List[Tuple[datetime.datetime, float]] = []
    sorted_data = sorted(self._data.items())

    start_time, _ = sorted_data[0]
    end_time, _ = sorted_data[-1]
    cur_start_time = start_time
    cur_end_time = cur_start_time + datetime.timedelta(seconds=window_size_sec)
    cur_total_count = 0
    for data_point in sorted_data:
      timestamp, count = data_point
      if timestamp >= cur_end_time:
        while timestamp >= cur_end_time:
          rates.append((cur_start_time, cur_total_count / window_size_sec))
          cur_start_time = cur_end_time
          cur_end_time = cur_start_time + datetime.timedelta(
              seconds=window_size_sec
          )
          cur_total_count = 0
      cur_total_count += count
    if cur_start_time <= end_time:
      delta_time_sec = (end_time - cur_start_time).total_seconds() + 1
      rates.append((cur_start_time, cur_total_count / delta_time_sec))
    return rates

  def rate_over_window_to_csv(self, window_size_sec: int) -> str:
    """Compute and return the rates over time and return them in csv string

    Args:
      window_size_sec: The size of the window in seconds for computing each
        individual rate

    Returns:
      A CSV string representation of the rates over time, with two rows:
      the first row contains timestamps, and the second row contains rate
      values.
    """
    rates = self.rate_over_window(window_size_sec)
    # Generate CSV string with two rows
    timestamps = "TimeStamp," + ",".join([str(e[0]) for e in rates])
    values = "Value," + ",".join([f"{e[1]:.2f}" for e in rates])
    csv_output = timestamps + "\n" + values
    return csv_output



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""mlperf loadgen interface for LLama2."""
import array
import concurrent.futures
import dataclasses
import json
import logging
from operator import itemgetter  # pylint: disable=g-importing-member
import time
from typing import List, Optional, Any

import numpy as np

import dataset

import mlperf_loadgen as lg

import grpc
from jetstream.core.proto import jetstream_pb2
from jetstream.core.proto import jetstream_pb2_grpc

from transformers import AutoTokenizer


logging.basicConfig(level=logging.INFO)
log = logging.getLogger("backend.py")

_last_log_time = 0  # Global variable to track last log time


def _log_once_n_sec(msg: str, log_interval_sec: int):
  """Logs a message once every specified number of seconds."""
  global _last_log_time
  current_time = time.time()
  if current_time - _last_log_time >= log_interval_sec:
    log.info("(once every %dsec): %s", log_interval_sec, msg)
    _last_log_time = current_time


@dataclasses.dataclass
class WarmupSample:
  id: int
  index: int


@dataclasses.dataclass
class StreamResponse:
  result: str = ""


class ThreadedLMClient:
  """Holds a thread pool and a loadgen client for LM inference."""

  _thread_pool: concurrent.futures.ThreadPoolExecutor
  _dataset: dataset.Dataset
  _futures = List[concurrent.futures.Future]

  def __init__(
      self,
      is_stream: bool,
      num_threads: int,
      api_url: str,
      dataset_object: dataset.Dataset,
      input_mode: str,
      output_mode: str,
      tokenizer: Optional[AutoTokenizer] = None,
      max_output_len: int = 1024,
      log_interval: int = 1000,
  ):
    log.info(f"Initiating {self.__class__.__name__} ...")
    self._is_stream = is_stream
    self._input_mode = dataset.validate_sample_mode(input_mode)
    self._output_mode = dataset.validate_sample_mode(output_mode)
    if self._input_mode == "text" or self._output_mode == "text":
      assert tokenizer is not None
    self._tokenizer = tokenizer
    self._max_output_len = max_output_len

    self._log_interval = log_interval

    self._thread_pool = concurrent.futures.ThreadPoolExecutor(num_threads)
    self._api_url = api_url
    self._dataset = dataset_object
    self._futures = []
    self.pred_outputs = {}
    self._resp_cnt = 0

    log.info("Creating grpc channel with api_url {}".format(api_url))
    options = [("grpc.keepalive_timeout_ms", 10000)]
    self._grpc_channel = grpc.insecure_channel(api_url, options=options)

  @property
  def tokenizer(self):
    return self._tokenizer

  def _log_resp_cnt(self):
    self._resp_cnt += 1
    if self._resp_cnt % self._log_interval == 0:
      log.info("Completed %d queries", self._resp_cnt)

  def process_single_sample_async(self, query_sample, warmup):
    """Executes a single query and marks responses complete asynchronously.

    Args:
      query_sample: Single prompt
      warmup: Indicates that this is a warmup request.
    """
    future = self._thread_pool.submit(
        self._process_sample, query_sample, warmup
    )
    self._futures.append(future)

  def flush(self):
    concurrent.futures.wait(self._futures)
    self._futures = []

  def _grpc_request(self, request, sample, warmup):
    """Send grpc synchronous request since the current grpc server is sync."""
    stub = jetstream_pb2_grpc.OrchestratorStub(self._grpc_channel)
    token_list = []
    ttft = 0
    start_time = time.perf_counter()
    response = stub.Decode(request)
    for resp in response:
      if not warmup and self._is_stream and ttft == 0:
        # TTFT for online mode
        ttft = time.perf_counter() - start_time
        _log_once_n_sec("TTFRT {}ms".format(ttft * 1000), 30)
        response_token_ids = resp.stream_content.samples[0].token_ids
        assert len(response_token_ids) == 1
        response_token_ids = np.array(response_token_ids, dtype=np.int64)
        response_array = array.array("B", response_token_ids.tobytes())
        response_info = response_array.buffer_info()
        first_token_response = lg.QuerySampleResponse(
            sample.id, response_info[0], response_info[1]
        )
        lg.FirstTokenComplete([first_token_response])
      token_list.extend(resp.stream_content.samples[0].token_ids)
    return token_list

  def _process_sample(self, sample, warmup):
    """Processes a single sample."""
    sample_data = self._dataset.inputs[sample.index]
    if self._input_mode == "text":
      token_ids = self._tokenizer.encode(sample_data)
    else:
      assert self._input_mode == "tokenized"
      token_ids = [int(token_id_str) for token_id_str in sample_data.split(",")]

    request = jetstream_pb2.DecodeRequest(
        token_content=jetstream_pb2.DecodeRequest.TokenContent(
            token_ids=token_ids
        ),
        max_tokens=self._max_output_len,
    )
    generated_token_list = self._grpc_request(request, sample, warmup)
    if not warmup:
      response_token_ids = generated_token_list
      n_tokens = len(response_token_ids)
      response_token_ids = np.array(response_token_ids, dtype=np.int64)
      response_array = array.array("B", response_token_ids.tobytes())
      response_info = response_array.buffer_info()
      response_data = response_info[0]
      response_size = response_info[1] * response_array.itemsize
      query_sample_response = lg.QuerySampleResponse(
          sample.id, response_data, response_size, n_tokens
      )
      lg.QuerySamplesComplete([query_sample_response])
      _log_once_n_sec("Mark query complete", 30)

      pred_output = self._tokenizer.decode(response_token_ids)
      self.pred_outputs[sample.index] = pred_output
      self._log_resp_cnt()


class SUT:
  """SUT."""

  def __init__(
      self,
      scenario,
      api_url,
      is_stream,
      input_mode,
      output_mode,
      max_output_len,
      dataset_path,
      total_sample_count,
      tokenizer_path=None,
      perf_count_override=None,
      num_client_threads=200,
      log_interval=1000,
      batch_size_exp=5,
      pred_outputs_log_path=None,
      dataset_rename_cols="",
  ):
    log.info(f"Starting {scenario} SUT with {api_url}.")
    self._is_stream = is_stream
    self._input_mode = dataset.validate_sample_mode(input_mode)
    self._output_mode = dataset.validate_sample_mode(output_mode)
    assert tokenizer_path is not None
    self._tokenizer = self.load_tokenizer(tokenizer_path)
    self._max_output_len = max_output_len
    self._api_url = api_url
    self._dataset_path = dataset_path
    self._total_sample_count = total_sample_count
    self._perf_count_override = perf_count_override
    self._num_client_threads = num_client_threads
    self._log_interval = log_interval
    self._batch_size_exp = batch_size_exp
    self._pred_outputs_log_path = pred_outputs_log_path

    log.info("Loading Dataset ... ")
    self.dataset = dataset.Dataset(
        dataset_path=self._dataset_path,
        input_mode=self._input_mode,
        total_sample_count=self._total_sample_count,
        perf_count_override=self._perf_count_override,
        dataset_rename_cols=dataset_rename_cols,
    )

    client_cls = ThreadedLMClient
    self._client = client_cls(
        is_stream=self._is_stream,
        num_threads=self._num_client_threads,
        api_url=self._api_url,
        dataset_object=self.dataset,
        input_mode=self._input_mode,
        output_mode=self._output_mode,
        tokenizer=self._tokenizer,
        max_output_len=self._max_output_len,
        log_interval=self._log_interval,
    )

    self.qsl = lg.ConstructQSL(
        self.dataset.total_sample_count,
        self.dataset.perf_count,
        self.dataset.LoadSamplesToRam,
        self.dataset.UnloadSamplesFromRam,
    )

    # We need to add some warmup to improve throughput estimation
    log.info("Starting warmup....")
    # Warm up with exponentially increasing batch sizes up to 32.
    for batch_size_exp in range(self._batch_size_exp):
      batch_size = 2**batch_size_exp
      for warmup_id, warmup_idx in enumerate(range(batch_size)):
        warmup_sample = WarmupSample(id=warmup_id, index=warmup_idx)
        self._client.process_single_sample_async(warmup_sample, True)
      self._client.flush()

    log.info("Warmup done....")
    time.sleep(30)
    self.sut = lg.ConstructSUT(self.issue_queries, self.flush_queries)

  def load_tokenizer(
      self, tokenizer_path: Optional[str] = None
  ) -> Optional[AutoTokenizer]:
    """Returns tokenizer"""
    if tokenizer_path is not None:
      tokenizer = AutoTokenizer.from_pretrained(
          tokenizer_path,
          model_max_length=1024,
          padding_side="left",
          use_fast=True,
      )
      tokenizer.pad_token = tokenizer.eos_token
      return tokenizer

  def _sort_issue_queries(self, query_samples):
    """Issue queries."""
    query_samples_with_length = []
    for query_sample in query_samples:
      query_sample_token_length = self.dataset.inputs_with_token_lengths[
          query_sample.index
      ][1]
      query_samples_with_length.append(
          (query_sample_token_length, query_sample)
      )
    sorted_query_samples_with_length = sorted(
        query_samples_with_length, key=itemgetter(0)
    )
    sorted_query_samples = [x[1] for x in sorted_query_samples_with_length]
    return sorted_query_samples

  def issue_queries(self, query_samples):
    """Issue queries."""
    num_query_samples = len(query_samples)
    if num_query_samples > 1:
      log.info(f"Issuing {num_query_samples} queries. ")
      query_samples = self._sort_issue_queries(query_samples)
    for query_sample in query_samples:
      self._client.process_single_sample_async(query_sample, False)

  def flush_queries(self):
    """Flush queries."""
    log.info("Loadgen has completed issuing queries... ")
    self._client.flush()

    if self._pred_outputs_log_path is not None:

      pred_outputs = []
      for idx, x in self._client.pred_outputs.items():
        pred_output = {
            "qsl_idx": idx,
            "intput": self._client._dataset.inputs[idx],
            "data": x,
        }
        pred_outputs.append(pred_output)
      log.info(f"Generated {len(pred_outputs)} prediction outputs")

      if pred_outputs:
        self.accuracy_log = open(self._pred_outputs_log_path, "w")
        self.accuracy_log.write(json.dumps(pred_outputs))
        self.accuracy_log.flush()
        self.accuracy_log.close()
        log.info("Dumpped prediction outputs to accuracy log... ")

  def __del__(self):
    print("Finished destroying SUT.")



"""
 Copyright 2025 Google LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

"""Recipe to test checkpointing and restore for Pathways and McJax.
 The recipe will run synchronously and will run the benchmark steps for each
 model and infrastructure combination. If TEST_RESTORE is set to True, it will
 restore from the initial run for the given number of steps and then continue
 the run.
"""

import datetime
import dataclasses
import sys
import os
import args_helper as helper

parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(parent_dir)

import maxtext_trillium_model_configs as model_configs
import maxtext_xpk_runner as mxr

PROXY_IMAGE = "us-docker.pkg.dev/cloud-tpu-v2-images-dev/pathways/unsanitized_proxy_server:latest"
SERVER_IMAGE = "us-docker.pkg.dev/cloud-tpu-v2-images-dev/pathways/unsanitized_server:latest"
RUNNER = "us-docker.pkg.dev/cloud-tpu-v2-images-dev/pathways/maxtext_jax_stable:latest"

# Cluster Params
CLUSTER = "v6e-256-cluster"
PROJECT = "tpu-prod-env-cluster"
ZONE = "us-east5-b"
COUNTRY = "us"
DEVICE_TYPE = "v6e-256"

# Other parameters (MUST BE SET BY USER)
XPK_PATH = "../xpk"  # We're running this script from the maxtext directory
USER = os.environ["USER"]
BASE_OUTPUT_DIRECTORY = (
    f"gs://{USER}-{PROJECT}-{COUNTRY}/pw_mcjax_benchmarking/"
)
# This needs to be set to True to test restore and if you want to restore from
# a previous run, you'll need to set RESUME_CHECKPOINT_NAMES below.
TEST_RESTORE = False
MAX_RESTARTS = 100

BENCHMARK_STEPS = 41
RESTORE_BENCHMARK_STEPS = 20  # Define steps for restore run

RESUME_CHECKPOINT_NAMES = {
    "pathways": {
        # Key is number of slices, value is a dictionary of run_name,
        # base_output_directory, and num_steps.
        32: {
            "run_name":  "restoring_run_name",
            "base_output_directory": f"gs://{BASE_OUTPUT_DIRECTORY}/...",
            "num_steps": BENCHMARK_STEPS + RESTORE_BENCHMARK_STEPS,
        }
    },
    # "mcjax": {
        # 32: {}
    # },
}


def _get_xpk_commands(
    models,
    cluster_config,
    pathways_config,
    num_steps=BENCHMARK_STEPS,
):
  """Generates xpk commands for the given models and configurations.

  Args:
    models: A dictionary of model lists, keyed by infrastructure type.
    cluster_config: The cluster configuration.
    num_slices_list: A list of the number of slices to use for each workload.
    pathways_config: The pathways configuration.
    num_steps: Number of steps for the workload.

  Returns:
    A list of tuples, where each tuple contains the xpk workload name, the
    xpk command, and the workload config.
  """
  xpk_workloads = []

  current_time = datetime.datetime.now()
  timestamp_str = current_time.strftime("%Y%m%d_%H%M%S")

  for infra, model_config in models.items():
    for model in model_config["models"]:
      # Run workloads on the below clusters
      for config in [
          cluster_config,
      ]:
        # Run workloads in the following slice configurations
        for num_slices in model_config["num_slices_list"]:
          # Ensure that we're checkpointing for the models.
          if model.pathways_tuning_params is None:
            model.pathways_tuning_params = model_configs.PATHWAYS_SHORT_RUN_CHECKPOINTING_TUNING_PARAMS

          run_name = None
          base_output_directory = (
              BASE_OUTPUT_DIRECTORY
              + f"{timestamp_str}_{infra}_{num_slices}_slice_{DEVICE_TYPE}_{model.model_name}/"
          )
          if infra in RESUME_CHECKPOINT_NAMES:
            if num_slices in RESUME_CHECKPOINT_NAMES[infra]:
              run_name = RESUME_CHECKPOINT_NAMES[infra][num_slices][
                  "run_name"
              ]
              base_output_directory = RESUME_CHECKPOINT_NAMES[infra][
                  num_slices
              ]["base_output_directory"]
              num_steps = RESUME_CHECKPOINT_NAMES[infra][num_slices][
                  "num_steps"
              ]

          wl_config = mxr.WorkloadConfig(
              model=model,
              num_slices=num_slices,
              device_type=config.device_type,
              base_output_directory=base_output_directory,
              max_restarts=MAX_RESTARTS,
              libtpu_type=None,
              libtpu_nightly_version="",
              base_docker_image=RUNNER if infra == "mcjax" else None,
              pathways_config=pathways_config if infra == "pathways" else None,
              xpk_path=XPK_PATH,
              num_steps=num_steps,
              priority="medium",
              run_name=run_name,
          )
          command, name = mxr.generate_xpk_workload_cmd(
              cluster_config=config, wl_config=wl_config, workload_name=run_name
          )

          print(f"Name of the workload is: {name} \n")
          print(f"XPK command to be used is: {command} \n")
          xpk_workloads.append(
              (name, command, wl_config)
          )

  return xpk_workloads


def _run_workloads_async(xpk_workloads, cluster_config, run_type="Initial"):
  """Runs the given xpk workloads asynchronously and yields workload names as they complete.

  Args:
    xpk_workloads: A list of tuples, each containing workload name, command and
      wl_config.
    cluster_config: The XPK cluster configuration.
    run_type: String to indicate if it's "Initial" or "Restore" run for logging.

  Yields:
    Tuple[workload_name, return_code, wl_config]: The name of the workload, its
      return code, and its workload config as it completes.
  """
  print(f"\nStarting {run_type} Workloads (Asynchronously)...\n")

  workload_configs_dict = {}  # Store wl_config by workload_name
  workload_names = []
  for workload_name, workload_cmd, wl_config in xpk_workloads:
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(
        f"[{timestamp}] Launching {run_type} workload: {workload_name} with"
        f" command: {workload_cmd}"
    )
    return_code = mxr.run_command_with_updates(
        workload_cmd, workload_name
    )

    if return_code != 0:
      print(
          f"Warning: Unable to start {run_type} xpk workload:"
          f" {workload_name}. Creation command failed, but continuing to"
          " launch others."
      )
      continue
    workload_names.append(workload_name)
    workload_configs_dict[workload_name] = wl_config # Store config

  # Wait for completion asynchronously and yield names and return codes as they
  # complete
  completed_workloads = mxr.wait_for_xpk_workloads_completion_async(
      cluster_config, workload_names, xpk_path=XPK_PATH
  )

  for completed_workload_name, return_code in completed_workloads:
    yield completed_workload_name, return_code, workload_configs_dict[
        completed_workload_name
    ]


def main() -> int:
  # V6e cluster config
  cluster_config = mxr.XpkClusterConfig(
      cluster_name=CLUSTER,
      project=PROJECT,
      zone=ZONE,
      device_type=DEVICE_TYPE,
  )

  # Handle command line arguments using args_helper
  should_continue = helper.handle_cmd_args(
      cluster_config, helper.DELETE, xpk_path=XPK_PATH
  )

  if not should_continue:
    return 0

  models = {
      "pathways": {
          "models": [
              model_configs.llama3_1_70b_8192_iter_synth_data_and_checkpointing,
          ],
          "num_slices_list": [
              2,
          ],
      },
      "mcjax": {
          "models": [
              model_configs.llama3_1_70b_8192_iter_synth_data_and_checkpointing,
          ],
          "num_slices_list": [
              2,
          ],
      },
  }
  pathways_config = mxr.PathwaysConfig(
      server_image=SERVER_IMAGE,
      proxy_server_image=PROXY_IMAGE,
      runner_image=RUNNER,

      # User can add additional flags here. We're adding StreamZ flags here.
      server_flags="--enable_metrics_collection=true",
      proxy_flags="--enable_metrics_collection=true",
      worker_flags="--enable_metrics_collection=true",
  )

  # --- Initial Run for Benchmark Steps ---
  print(
      "\n--- Starting Initial Benchmark Run ---"
  )
  xpk_workloads_initial = _get_xpk_commands(
      models,
      cluster_config,
      pathways_config,
      num_steps=BENCHMARK_STEPS,
  )

  completed_initial_workloads = {}
  initial_workload_configs = {}

  # Iterate through completed workloads as they yield
  for completed_workload_name, return_code, wl_config in _run_workloads_async(
      xpk_workloads_initial, cluster_config, run_type="Initial"
  ):
    print(
        f"\n--- Initial Workload '{completed_workload_name}' COMPLETED with"
        f" code: {return_code}. ---\n"
    )
    if return_code == 0:
      completed_initial_workloads[completed_workload_name] = (
          return_code  # Track completed workload names
      )
      initial_workload_configs[completed_workload_name] = (
          wl_config  # Store the config
      )
    else:
      print(
          f"--- Workload '{completed_workload_name}' FAILED, NOT restoring. ---"
      )

    if TEST_RESTORE and return_code == 0:
      # --- Restore Run for Additional Steps ---
      print(f"\n--- Starting Restore Run for '{completed_workload_name}' ---")

      # First delete the workload so we can restore it from scratch.
      helper.handle_delete_specific_workload(
          cluster_config, completed_workload_name, xpk_path=XPK_PATH
      )

      original_wl_config = initial_workload_configs[completed_workload_name]

      # Create new config for restore with the number of steps increased
      restore_wl_config = dataclasses.replace(
          original_wl_config,
          num_steps=original_wl_config.num_steps + RESTORE_BENCHMARK_STEPS,
      )

      restore_command, _ = mxr.generate_xpk_workload_cmd(
          cluster_config=cluster_config,
          wl_config=restore_wl_config,
          workload_name=completed_workload_name,
      )
      print(
          f"Restore command for '{completed_workload_name}': {restore_command}"
      )

      print(f"Launching restore workload: {completed_workload_name}")
      restore_return_code = mxr.run_command_with_updates(
          restore_command, f"Restore {completed_workload_name}"
      )
      if restore_return_code == 0:
        print(
            f"\n--- Restore Workload for '{completed_workload_name}' launched"
            " successfully. ---"
        )
      else:
        print(
            f"\n--- Restore Workload for '{completed_workload_name}' FAILED to"
            " launch. ---"
        )

  return 0


if __name__ == "__main__":
  main()
